#include <iostream>
#include <fstream>
#include <algorithm>
using namespace std;
/*
	This program automatically patches your slope output to your map without manual hex editing.
	Make sure DSlopeOut_Map(yourmap).txt exists in the luafiles directory.
*/
void stringclear(string &str) {
	//Removes quotations from file path.
	str.erase(
	    remove( str.begin(), str.end(), '\"' ),
	    str.end()
    );
}

string getFileExt(const string& s) {

   size_t i = s.rfind('.', s.length());
   if (i != string::npos) {
      return(s.substr(i+1, s.length() - i));
   }

   return("");
}

uint32_t bufgetpos(string str, uint32_t pos, string marker){
	uint32_t addr = *reinterpret_cast<uint32_t *>(&str[str.find(marker,pos)-8]);
	return addr;	
}
int main() {
	fstream wad;
	string wadpath;
	string patchpath;
	ifstream patch;
    std::cout << "Input wad file path: ";
  	std::getline(std::cin, wadpath);
    std::cout << "Input wad patch path: ";
  	std::getline(std::cin, patchpath);
  	
	stringclear(wadpath);
	stringclear(patchpath);
  	wad.open(wadpath, std::ios::out | std::ios::in| std::ios_base::binary);
  	patch.open(patchpath,ios::in);
  	if (patch.fail() || wad.fail()){
  			cout<<"Invalid Files."<<endl;
		  return -1;
	  }
	if( getFileExt(wadpath)  != "wad" || getFileExt(patchpath) != "txt"){
		cout<<"The first file must be a .wad and the second a .txt."<<endl;
		return -1;
	}
	
  	wad.seekg(0,wad.end);
	int length = wad.tellg();
  	wad.seekg(0,wad.beg);
  	
  	string wadbuffer(length,'\0');
  	wad.read(&wadbuffer[0], length);
  	wad.close();
  	//cout << length <<endl;
  	//Find a marker in wad equivalent to the needed map
  	string mx = patchpath.substr(patchpath.find_last_of('_')+1);
  	string mapname = mx.erase(mx.find_last_of("."));
  	cout << mapname <<endl;
  	//Find starting marker, then find next instances of SECTORS, LINEDEFS, SIDEDEFS markers after starting position
	ofstream out(wadpath,std::ios_base::out | std::ios_base::binary);
	//Make a backup
	ofstream back(wadpath += ".bak",std::ios_base::out | std::ios_base::binary);
	back.write(&wadbuffer[0],wadbuffer.size());
	uint32_t mapmarker = wadbuffer.find(mapname);
	if(wadbuffer.find(mapname) == string::npos){
		cout<<"Invalid map or patch file name."<<endl;
		return -1;
	}
  	uint32_t linemarker = bufgetpos(wadbuffer, mapmarker, "LINEDEFS");
  	uint32_t ssectormarker = wadbuffer.find("SSECTORS",mapmarker);	//SSECTORS needs to be known in order to ignore it when looking for SECTORS address.
  	uint32_t sectormarker = bufgetpos(wadbuffer, ssectormarker+8, "SECTORS");
  	uint32_t sidefmarker = bufgetpos(wadbuffer, mapmarker, "SIDEDEFS");
  	//cout<<"Lines: "<<wadbuffer.find("LINEDEFS",mapmarker)<<'\n';  
	//cout<<linemarker<<" "<<sectormarker<<" "<<sidefmarker<<'\n';
	
	//determine what base offset to use by #1 (sector), #2 (line), or #3 (sidedef)
	string patchline;
	uint32_t basemarker;
	uint32_t offsetmarker;
	uint32_t consoffset = 0;
	uint32_t markersize = 0;
	char bytedata[60]{};
	while(getline(patch,patchline)) {
		transform(patchline.begin(), patchline.end(), patchline.begin(), ::toupper);
		if (patchline[0] == '#' ){
			switch (patchline[1]) {
				case '1':
					basemarker = sectormarker;
					markersize = 26;
					break;
				case '2':
					basemarker = linemarker;
					markersize = 14;
					break;
				case '3':
					basemarker = sidefmarker;
					markersize = 30;
					break;
			}
		}
		else if (patchline[0] == '.') {
			consoffset = 0;
			offsetmarker = stoull(patchline.substr(1,string::npos),nullptr,16);
		}
		else{
			if(patchline.length() > 0){
			    for(int j = 0; j < markersize; j++) {
			        sscanf(&patchline[0] + j * 2, "%02hhX", bytedata + j);
			    }
				if(basemarker == sidefmarker){
					for(int i = 0 ; i<4; i++){
			    		wadbuffer[basemarker+offsetmarker+(consoffset*markersize)+i] = bytedata[i];
					}	
				}
				else {
					 for(int i = 0 ; i<markersize; i++){
			    		wadbuffer[basemarker+offsetmarker+(consoffset*markersize)+i] = bytedata[i];
					}	
				}
				consoffset++;
			}
		}
	}
	//wad.write(&wadbuffer[0],wadbuffer.size());
  	out.write(&wadbuffer[0],wadbuffer.size());
  	cout<<"Patch Complete!"<<endl;
	patch.close();
  	out.close();
    return 0;
}
